/* -*- Mode: C; indent-tabs: t; c-basic-offset: 9; tab-width: 9 -*- */
/*
 *  This file is part of itouch-ctrl plugin for audacious
 *
 *  Copyright (c) 2006 Vladimir Paskov <vlado.paskov@gmail.com>
 *  Name: plugin.c
 *  Description: plugin.c
 * 
 *  Part of this code are from xmms-itouch plugin.
 *  Authors of xmms-itouch are listed below:
 *
 *  Copyright (C) 2000-2002 Ville Syrjälä <syrjala@sci.fi>
 *                         Bryn Davies <curious@ihug.com.au>
 *                         Jonathan A. Davis <davis@jdhouse.org>
 *                         Jeremy Tan <nsx@nsx.homeip.net>
 *
 *  itouch-ctrl is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  itouch-ctrl is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with itouch-ctrl; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include "i18n.h"
#include <stdlib.h>
#include <X11/Xlib.h>
#include <gtk/gtk.h>
#include <audacious/plugin.h>
#include <audacious/beepctrl.h>
#include <audacious/configdb.h>

#define PLUG_FNAME "libitouchctrl.la"
#define PLUG_DESC "iTouch Control"

/* func defs */
void x_display_init (void);
static void init (void);
static void itouchctrl_grab_keys (Display *display, Window window);
static void itouchctrl_ungrab_keys (Display *display, Window window);
static gint itouchctrl_timeout_func(gpointer data);
static void itouchctrl_handle_keyevent(int keycode);
static void itouchctrl_load_config (void);
static void itouchctrl_save_config (void);
static void configure (void);
static void itouchctrl_grab_keyboard (GtkWidget *widget, gpointer data);
void itouchctrl_cancel_callback (GtkWidget *widget);
void itouchctrl_ok_callback (GtkWidget *widget, gpointer data);
static void about (void);
static void cleanup (void);

typedef struct {
	gint vol_increment;
	gint vol_decrement;

	gint mute;
	gint vol_down;
	gint vol_up;
	gint play;
	gint stop;
	gint prev_track;
	gint next_track;
} iTouchConfig;

iTouchConfig itouch_cfg;

typedef struct {
	gchar *name;
	gint mute;
	gint vol_down;
	gint vol_up;
	gint play;
	gint stop;
	gint prev_track;
	gint next_track;
} KeyboardConfig;

KeyboardConfig *kbdcfgs = NULL;

static Display *xdisplay;
static Window x_root_window;
static gint timeout_tag = 0;
gint grabbed = 0;
gint init_done = 0;
gboolean loaded = FALSE;

typedef struct {
	GtkWidget *play_spin;
	GtkWidget *stop_spin;
	GtkWidget *prev_spin;
	GtkWidget *next_spin;
	GtkWidget *up_spin;
	GtkWidget *down_spin;
	GtkWidget *mute_spin;
} iTouchSpin;

static GeneralPlugin itouchctrl =
{
	NULL,
	NULL,
	-1,
	PLUG_DESC,
	init,
	about,
	configure,
	cleanup
};

GeneralPlugin *get_gplugin_info (void)
{
	itouchctrl.description = 
	g_strdup_printf (_("iTouchControl %s"), VERSION);
	
	return &itouchctrl;
}

/* 
 * plugin init start 
 */
static void init (void)
{
	/* check X display */
	x_display_init ( );
	
	timeout_tag = gtk_timeout_add (100, itouchctrl_timeout_func, NULL);
	
	/* load plugin configuration */
	itouchctrl_load_config ( );
	
	/* grab keys */
	itouchctrl_grab_keys (xdisplay, x_root_window);

	loaded = TRUE;
}

/* check X display */
void x_display_init (void)
{
	gchar *display_name = (gchar*)getenv ("DISPLAY");
	
	if (!display_name)
	{
		perror(_("No display\n"));
		return;
	}
	
	xdisplay = XOpenDisplay (display_name);
	if (!xdisplay)
	{
		perror(_("Can't open display\n"));
		return;
	}
	
	x_root_window = DefaultRootWindow (xdisplay);
	if (!x_root_window)
	{
		perror(_("No root window\n"));
		return;
	}
	
	init_done = 1;
}

static gint itouchctrl_timeout_func (gpointer data)
{
	long mask = KeyPressMask;
	XEvent event;
	
	while (XCheckMaskEvent(xdisplay, mask, &event))
	{
		itouchctrl_handle_keyevent (event.xkey.keycode);
	}
	
	return TRUE;
}

static void itouchctrl_handle_keyevent (int keycode)
{
	gint current_volume, old_volume;
	gboolean play, mute;
	
	/* playing or not */
	play = xmms_remote_is_playing (itouchctrl.xmms_session);
	
	/* get current volume */
	current_volume = xmms_remote_get_main_volume (itouchctrl.xmms_session);
	
	if (current_volume)
	{
		mute = FALSE;
	} else {
		mute = TRUE;
	}
	
	if (keycode == itouch_cfg.mute)
	{
		if (!mute)
		{
			current_volume = old_volume;
			current_volume = 0;
			mute = TRUE;
		} else {
			current_volume = old_volume;
			mute = FALSE;
		}
			
		xmms_remote_set_main_volume (itouchctrl.xmms_session,
					current_volume);
		return;
	}
		
	if (keycode == itouch_cfg.vol_down)
	{
		if (mute)
		{
			current_volume = old_volume;
			old_volume = 0;
			mute = FALSE;
		}
			
		if ((current_volume -= itouch_cfg.vol_decrement) < 0)
		{
			current_volume = 0;
		}
			
		if (current_volume != old_volume)
		{
			xmms_remote_set_main_volume (itouchctrl.xmms_session,
						current_volume);
		}
			
		old_volume = current_volume;
		return;
	}
		
	if (keycode == itouch_cfg.vol_up)
	{
		if (mute)
		{
			current_volume = old_volume;
			old_volume = 0;
			mute = FALSE;
		}
			
		if ((current_volume += itouch_cfg.vol_increment) > 100)
		{
			current_volume = 100;
		}
			
		if (current_volume != old_volume)
		{
			xmms_remote_set_main_volume (itouchctrl.xmms_session,
						current_volume);
		}
			
		old_volume = current_volume;
		return;
	}
		
	if (keycode == itouch_cfg.play)
	{
		if (!play)
		{
			xmms_remote_play (itouchctrl.xmms_session);
		} else {
			xmms_remote_pause (itouchctrl.xmms_session);
		}
			
		return;
	}
		
	if (keycode == itouch_cfg.stop)
	{
		if (play)
		{
			xmms_remote_stop (itouchctrl.xmms_session);
		} else {
			xmms_remote_eject (itouchctrl.xmms_session);
		}
			
		return;
	}
		
	if (keycode == itouch_cfg.prev_track)
	{
		xmms_remote_playlist_prev (itouchctrl.xmms_session);
		return;
	}
		
	if (keycode == itouch_cfg.next_track)
	{
		xmms_remote_playlist_next (itouchctrl.xmms_session);
		return;
	}
}

/* load plugin configuration */
static void itouchctrl_load_config (void)
{
	ConfigDb *cfdb;

	/* default volume level */
	itouch_cfg.vol_increment = 4;
	itouch_cfg.vol_decrement = 4;

	itouch_cfg.mute = 0;
	itouch_cfg.vol_down = 0;
	itouch_cfg.vol_up = 0;
	itouch_cfg.play = 0;
	itouch_cfg.stop = 0;
	itouch_cfg.prev_track = 0;
	itouch_cfg.next_track = 0;

	/* open configuration database */
	cfdb = bmp_cfg_db_open ( );

	bmp_cfg_db_get_int (cfdb, "iTouch", "mute", &itouch_cfg.mute);
	bmp_cfg_db_get_int (cfdb, "iTouch", "vol_down", &itouch_cfg.vol_down);
	bmp_cfg_db_get_int (cfdb, "iTouch", "vol_up", &itouch_cfg.vol_up);
	bmp_cfg_db_get_int (cfdb, "iTouch", "play", &itouch_cfg.play);
	bmp_cfg_db_get_int (cfdb, "iTouch", "stop", &itouch_cfg.stop);
	bmp_cfg_db_get_int (cfdb, "iTouch", "prev_track", &itouch_cfg.prev_track);
	bmp_cfg_db_get_int (cfdb, "iTouch", "next_track", &itouch_cfg.next_track);

	bmp_cfg_db_close (cfdb);
}

/* save plugin configuration */
static void itouchctrl_save_config (void)
{
	ConfigDb *cfdb;
	
	/* open configuration database */
	cfdb = bmp_cfg_db_open ( );
	
	bmp_cfg_db_set_int (cfdb, "iTouch", "mute", itouch_cfg.mute);
	bmp_cfg_db_set_int (cfdb, "iTouch", "vol_up", itouch_cfg.vol_up);
	bmp_cfg_db_set_int (cfdb, "iTouch", "vol_down", itouch_cfg.vol_down);
	bmp_cfg_db_set_int (cfdb, "iTouch", "play", itouch_cfg.play);
	bmp_cfg_db_set_int (cfdb, "iTouch", "stop", itouch_cfg.stop);
	bmp_cfg_db_set_int (cfdb, "iTouch", "prev_track", itouch_cfg.prev_track);
	bmp_cfg_db_set_int (cfdb, "iTouch", "next_track", itouch_cfg.next_track);
	
	bmp_cfg_db_close (cfdb);
}

/* grab requied keys */
static void itouchctrl_grab_keys (Display *display, Window window)
{
	if (grabbed) return;
	
	if (itouch_cfg.mute)
	{
		XGrabKey (display, itouch_cfg.mute, AnyModifier, x_root_window, 
			False, GrabModeAsync, GrabModeAsync);
	}
	
	if (itouch_cfg.vol_up)
	{
		XGrabKey (display, itouch_cfg.vol_up, AnyModifier, x_root_window, 
			False, GrabModeAsync, GrabModeAsync);
	}

	if (itouch_cfg.vol_down)
	{
		XGrabKey (display, itouch_cfg.vol_down, AnyModifier, x_root_window, 
			False, GrabModeAsync, GrabModeAsync);
	}
	
	if (itouch_cfg.play)
	{
		XGrabKey (display, itouch_cfg.play, AnyModifier, x_root_window, 
			False, GrabModeAsync, GrabModeAsync);
	}

	if (itouch_cfg.stop)
	{
		XGrabKey (display, itouch_cfg.stop, AnyModifier, x_root_window, 
			False, GrabModeAsync, GrabModeAsync);
	}
	
	if (itouch_cfg.prev_track)
	{
		XGrabKey (display, itouch_cfg.prev_track, AnyModifier, x_root_window, 
			False, GrabModeAsync, GrabModeAsync);
	}
	
	if (itouch_cfg.next_track)
	{
		XGrabKey (display, itouch_cfg.next_track, AnyModifier, x_root_window, 
			False, GrabModeAsync, GrabModeAsync);
	}
	
	grabbed = 1;
}
/*
 * plugin init end
 */

static void configure (void)
{
	iTouchSpin *ispin;
	GtkWidget *window;
	GtkWidget *main_vbox, *vbox;
	GtkWidget *hbox;
	GtkWidget *alignment;
	GtkWidget *frame;
	GtkWidget *label;
	GtkWidget *image;
	GtkWidget *table;
	GtkWidget *grab_button;
	GtkWidget *button_box, *button;
	
	itouchctrl_ungrab_keys(xdisplay, x_root_window);
	
	ispin = (iTouchSpin*)g_malloc(sizeof(iTouchSpin));
	if (!ispin)
	{
		return;
	}
	
	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW (window), _("iTouch Control Plugin Configuration"));
	gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER_ALWAYS);
	gtk_window_set_type_hint (GTK_WINDOW (window), GDK_WINDOW_TYPE_HINT_DIALOG);
	gtk_window_set_resizable (GTK_WINDOW (window), FALSE);
	gtk_container_set_border_width (GTK_CONTAINER (window), 2);
	
	main_vbox = gtk_vbox_new (FALSE, 4);
	gtk_container_add (GTK_CONTAINER (window), main_vbox);
	
	alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
	gtk_box_pack_start (GTK_BOX (main_vbox), alignment, FALSE, TRUE, 0);
	gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 4, 0, 0, 0);
	hbox = gtk_hbox_new (FALSE, 2);
	gtk_container_add (GTK_CONTAINER (alignment), hbox);
	image = gtk_image_new_from_stock (GTK_STOCK_DIALOG_INFO, GTK_ICON_SIZE_DIALOG);
	gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, TRUE, 0);
	label = gtk_label_new (_("Type the keycode of the multimedia key or click\n"
				"the \"Grab\" button and then press the key."));
	gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	
	label = gtk_label_new (NULL);
	gtk_label_set_markup (GTK_LABEL (label), _("<b>Playback:</b>"));
	frame = gtk_frame_new (NULL);
	gtk_frame_set_label_widget (GTK_FRAME (frame), label);
	gtk_box_pack_start (GTK_BOX (main_vbox), frame, FALSE, TRUE, 0);
	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
	alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
	gtk_container_add (GTK_CONTAINER (frame), alignment);
	gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 3, 3, 3, 3);
	vbox = gtk_vbox_new (FALSE, 2);
	gtk_container_add (GTK_CONTAINER (alignment), vbox);
	label = gtk_label_new (NULL);
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, TRUE, 0);
	gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_CENTER);
	gtk_misc_set_alignment (GTK_MISC (label), 0.5, 0.5);
	gtk_label_set_markup (GTK_LABEL (label), 
			_("<i>Configure multimedia keys wich controls Audacious playback.</i>"));
	table = gtk_table_new (4, 3, FALSE);
	gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);
	gtk_table_set_col_spacings (GTK_TABLE (table), 2);
	gtk_table_set_row_spacings (GTK_TABLE (table), 2);

	/* play */
	label = gtk_label_new (_("Play:"));
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1, 
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_misc_set_padding (GTK_MISC (label), 3, 3);
	
	ispin->play_spin = gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 300, 1, 5, 0)), 1, 0);
	gtk_table_attach (GTK_TABLE (table), ispin->play_spin, 1, 2, 0, 1,
			(GtkAttachOptions) (GTK_FILL|GTK_EXPAND), (GtkAttachOptions) (GTK_EXPAND), 0, 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (ispin->play_spin), itouch_cfg.play);
	
	grab_button = gtk_button_new_with_label (_("Grab"));
	gtk_table_attach (GTK_TABLE (table), grab_button, 2, 3, 0, 1, 
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	g_signal_connect (G_OBJECT (grab_button), "clicked",
			G_CALLBACK (itouchctrl_grab_keyboard), ispin->play_spin);

	/* stop */
	label = gtk_label_new (_("Stop:"));
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2, 
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_misc_set_padding (GTK_MISC (label), 3, 3);
	
	ispin->stop_spin = gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 300, 1, 5, 0)), 1, 0);
	gtk_table_attach (GTK_TABLE (table), ispin->stop_spin, 1, 2, 1, 2, 
			(GtkAttachOptions) (GTK_FILL|GTK_EXPAND), (GtkAttachOptions) (GTK_EXPAND), 0, 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (ispin->stop_spin), itouch_cfg.stop);
	
	grab_button = gtk_button_new_with_label (_("Grab"));
	gtk_table_attach (GTK_TABLE (table), grab_button, 2, 3, 1, 2, (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	g_signal_connect (G_OBJECT (grab_button), "clicked",
			G_CALLBACK (itouchctrl_grab_keyboard), ispin->stop_spin);

	/* prev track */
	label = gtk_label_new (_("Previous Track:"));
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, 2, 3, 
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_misc_set_padding (GTK_MISC (label), 3, 3);
	
	ispin->prev_spin = gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 300, 1, 5, 0)), 1, 0);
	gtk_table_attach (GTK_TABLE (table), ispin->prev_spin, 1, 2, 2, 3, 
			(GtkAttachOptions) (GTK_FILL|GTK_EXPAND), (GtkAttachOptions) (GTK_EXPAND), 0, 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (ispin->prev_spin), itouch_cfg.prev_track);
	
	grab_button = gtk_button_new_with_label (_("Grab"));
	gtk_table_attach (GTK_TABLE (table), grab_button, 2, 3, 2, 3, (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	g_signal_connect (G_OBJECT (grab_button), "clicked",
			G_CALLBACK (itouchctrl_grab_keyboard), ispin->prev_spin);

	/* next track */
	label = gtk_label_new (_("Next Track:"));
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, 3, 4, 
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_misc_set_padding (GTK_MISC (label), 3, 3);
	
	ispin->next_spin = gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 300, 1, 5, 0)), 1, 0);
	gtk_table_attach (GTK_TABLE (table), ispin->next_spin, 1, 2, 3, 4, 
			(GtkAttachOptions) (GTK_FILL|GTK_EXPAND), (GtkAttachOptions) (0), 0, 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (ispin->next_spin), itouch_cfg.next_track);
	
	grab_button = gtk_button_new_with_label (_("Grab"));
	gtk_table_attach (GTK_TABLE (table), grab_button, 2, 3, 3, 4, 
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	g_signal_connect (G_OBJECT (grab_button), "clicked",
			G_CALLBACK (itouchctrl_grab_keyboard), ispin->next_spin);

	label = gtk_label_new (NULL);
	gtk_label_set_markup (GTK_LABEL (label), _("<b>Volume Control:</b>"));
	frame = gtk_frame_new (NULL);
	gtk_frame_set_label_widget (GTK_FRAME (frame), label);
	gtk_box_pack_start (GTK_BOX (main_vbox), frame, FALSE, TRUE, 0);
	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
	alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
	gtk_container_add (GTK_CONTAINER (frame), alignment);
	gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 3, 3, 3, 3);
	vbox = gtk_vbox_new (FALSE, 2);
	gtk_container_add (GTK_CONTAINER (alignment), vbox);
	label = gtk_label_new (NULL);
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, TRUE, 0);
	gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_CENTER);
	gtk_misc_set_alignment (GTK_MISC (label), 0.5, 0.5);
	gtk_label_set_markup (GTK_LABEL (label), 
			_("<i>Configure multimedia keys wich controls music volume.</i>"));
	table = gtk_table_new (3, 3, FALSE);
	gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);
	gtk_table_set_col_spacings (GTK_TABLE (table), 2);
	gtk_table_set_row_spacings (GTK_TABLE (table), 2);
	
	/* mute */
	label = gtk_label_new (_("Mute:"));
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1, 
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_misc_set_padding (GTK_MISC (label), 3, 3);
	
	ispin->mute_spin = gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 300, 1, 5, 0)), 1, 0);
	gtk_table_attach (GTK_TABLE (table), ispin->mute_spin, 1, 2, 0, 1,
			(GtkAttachOptions) (GTK_FILL|GTK_EXPAND), (GtkAttachOptions) (GTK_EXPAND), 0, 0);
	
	grab_button = gtk_button_new_with_label (_("Grab"));
	gtk_table_attach (GTK_TABLE (table), grab_button, 2, 3, 0, 1, (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	g_signal_connect (G_OBJECT (grab_button), "clicked",
			G_CALLBACK (itouchctrl_grab_keyboard), ispin->mute_spin);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (ispin->mute_spin), itouch_cfg.mute);
	
	/* volume up */
	label = gtk_label_new (_("Volume Up:"));
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2, 
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_misc_set_padding (GTK_MISC (label), 3, 3);
	
	ispin->up_spin = gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 300, 1, 5, 0)), 1, 0);
	gtk_table_attach (GTK_TABLE (table), ispin->up_spin, 1, 2, 1, 2, 
			(GtkAttachOptions) (GTK_FILL|GTK_EXPAND), (GtkAttachOptions) (GTK_EXPAND), 0, 0);
	
	grab_button = gtk_button_new_with_label (_("Grab"));
	gtk_table_attach (GTK_TABLE (table), grab_button, 2, 3, 1, 2, 
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	g_signal_connect (G_OBJECT (grab_button), "clicked",
			G_CALLBACK (itouchctrl_grab_keyboard), ispin->up_spin);

	/* volume down */
	label = gtk_label_new (_("Volume Down:"));
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, 2, 3, 
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_misc_set_padding (GTK_MISC (label), 3, 3);
	
	ispin->down_spin = gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 300, 1, 5, 0)), 1, 0);
	gtk_table_attach (GTK_TABLE (table), ispin->down_spin, 1, 2, 2, 3, 
			(GtkAttachOptions) (GTK_FILL|GTK_EXPAND), (GtkAttachOptions) (GTK_EXPAND), 0, 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (ispin->down_spin), itouch_cfg.vol_down);
	
	grab_button = gtk_button_new_with_label (_("Grab"));
	gtk_table_attach (GTK_TABLE (table), grab_button, 2, 3, 2, 3, 
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	g_signal_connect (G_OBJECT (grab_button), "clicked",
			G_CALLBACK (itouchctrl_grab_keyboard), ispin->down_spin);
	
	button_box = gtk_hbutton_box_new ( );
	gtk_box_pack_start (GTK_BOX (main_vbox), button_box, FALSE, TRUE, 0);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (button_box), GTK_BUTTONBOX_END);
	gtk_box_set_spacing (GTK_BOX (button_box), 4);
	
	button = gtk_button_new_from_stock (GTK_STOCK_CANCEL);
	gtk_container_add (GTK_CONTAINER (button_box), button);
	g_signal_connect (G_OBJECT (button), "clicked",
			G_CALLBACK (itouchctrl_cancel_callback), window);
	
	button = gtk_button_new_from_stock (GTK_STOCK_OK);
	gtk_container_add (GTK_CONTAINER (button_box), button);
	g_signal_connect (G_OBJECT (button), "clicked",
			G_CALLBACK (itouchctrl_ok_callback), ispin);
	
	gtk_widget_show_all (GTK_WIDGET (window));
}

static void about (void)
{
	static GtkWidget *dialog;
	
	dialog = xmms_show_message (_("About iTouch Control Plugin"),
				_("iTouch Control Plugin\n\n"
				"Copyright (C) 2006 Vladimir Paskov <vlado.paskov@gmail.com>\n\n"
				"Parts of the plugin source are from xmms-itouch plugin.\n"
				"Authors of xmms-itouch are listed below:\n"
				"Copyright (C) 2000-2002 Ville Syrjälä <syrjala@sci.fi>\n"
                         	"			Bryn Davies <curious@ihug.com.au>\n"
                        	"			Jonathan A. Davis <davis@jdhouse.org>\n"
                         	"			Jeremy Tan <nsx@nsx.homeip.net>\n\n"
                         	"For questions, feature request and bug reports:\n"
                         	"http://sourceforge.net/projects/itouch-control"),
                         	_("Ok"), FALSE, NULL, NULL);
                         	
	gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
			   GTK_SIGNAL_FUNC(gtk_widget_destroyed), &dialog);						
}

/* grab key callback */
static void itouchctrl_grab_keyboard (GtkWidget *widget, gpointer data)
{
	GtkWidget *spin = GTK_WIDGET (data);
	long mask = KeyPressMask;
	XEvent event;
	
	if (!init_done)
	{
		x_display_init ( );
	}
	
	XGrabKeyboard (xdisplay, x_root_window, False, GrabModeAsync, GrabModeAsync, CurrentTime);
	XMaskEvent(xdisplay, mask, &event);

	XUngrabKeyboard(xdisplay, CurrentTime);

	gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin), event.xkey.keycode);
}

void itouchctrl_cancel_callback (GtkWidget *widget)
{
	if (loaded)
	{
		itouchctrl_grab_keys (xdisplay, x_root_window);
	}

	gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (widget)));
}

void itouchctrl_ok_callback (GtkWidget *widget, gpointer data)
{
	iTouchSpin *ispin = (iTouchSpin*)data;
	
	itouch_cfg.play =
	gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (ispin->play_spin));
	
	itouch_cfg.stop =
	gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (ispin->stop_spin));
	
	itouch_cfg.prev_track = 
	gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (ispin->prev_spin));
	
	itouch_cfg.next_track =
	gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (ispin->next_spin));
	
	itouch_cfg.vol_up =
	gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (ispin->up_spin));
	
	itouch_cfg.vol_down = 
	gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (ispin->down_spin));
	
	itouchctrl_save_config ( );
	
	if (loaded)
	{
		itouchctrl_grab_keys (xdisplay, x_root_window);
	}
	
	gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (widget)));
}

/* 
 * plugin cleanup 
 */
 static void cleanup (void)
 {
 	itouchctrl_ungrab_keys (xdisplay, x_root_window);
 	
 	if (timeout_tag)
 	{
 		gtk_timeout_remove(timeout_tag);
 	}
 	
 	timeout_tag = 0;
	XCloseDisplay(xdisplay);
	loaded = FALSE;
 }
 
 static void itouchctrl_ungrab_keys (Display *display, Window window)
 {
 	if (!grabbed) return;
 	
 	if (itouch_cfg.mute)
 	{
 		XUngrabKey(display, itouch_cfg.mute, AnyModifier, window);
 	}
 	
 	if (itouch_cfg.vol_up)
 	{
 		XUngrabKey(display, itouch_cfg.vol_up, AnyModifier, window);
 	}
 	
 	if (itouch_cfg.vol_down)
 	{
 		XUngrabKey(display, itouch_cfg.vol_down, AnyModifier, window);
 	}
 	
 	if (itouch_cfg.play)
 	{
 		XUngrabKey(display, itouch_cfg.play, AnyModifier, window);
 	}
 	
 	if (itouch_cfg.stop)
 	{
 		XUngrabKey(display, itouch_cfg.stop, AnyModifier, window);
 	}
 	
 	if (itouch_cfg.prev_track)
 	{
 		XUngrabKey(display, itouch_cfg.prev_track, AnyModifier, window);
 	}
 	
 	if (itouch_cfg.next_track)
 	{
 		XUngrabKey(display, itouch_cfg.next_track, AnyModifier, window);
 	}
 	
 	grabbed = 0;
}

